/*
 *  tcp.c - functions logging TCP connections
 *
 *  Copyright (C) 1998-1999 Hugo Haas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>

#ifdef _MULTITHREAD_
#include <pthread.h>
#endif

#include "defines.h"
#include "tcp.h"
#include "netutils.h"
#include "log.h"
#include "filter.h"
#include "configuration.h"
#include "ident.h"

extern struct logger log;

#define SERVICE_LENGTH 512

extern char *configuration_file;

/*
 * Structure of a TCP packet
 */

struct tcppacket {
  struct iphdr ip;
  struct tcphdr tcp;
};

/*
 * service_lookup
 *
 * Get a service name
 */

void tcp_service_lookup(char *service, __u16 port) {
  struct servent *se;

  se = getservbyport(port, "tcp");
  if (se == NULL)
    sprintf(service, "port %d", ntohs(port));
  else {
    sprintf(service, "%s", se->s_name);
    service[SERVICE_LENGTH-1] = '\0';
  }
}

/*
 * log_tcppacket & threaded_log_tcppacket
 *
 * Log a TCP connection
 */

void *log_tcppacket(void *pkt) {
  char remote_host[HOST_LENGTH];
  char service[SERVICE_LENGTH];
#if 0
  char username[USERNAME_LENGTH];
#endif

  host_lookup(remote_host, ((struct tcppacket *) pkt)->ip.saddr);
  /* Host filter */
  if (do_log(((struct tcppacket *) pkt)->ip.saddr, remote_host, ((struct tcppacket *) pkt)->tcp.dest, TCP_PROTO) == TRUE) {
    tcp_service_lookup(service, ((struct tcppacket *) pkt)->tcp.dest);
#if 0
    if (((struct tcppacket *) pkt)->tcp.dest != TCP_IDENT_PORT) {
      get_ident_info(username, ntohs(((struct tcppacket *) pkt)->tcp.source), ntohs(((struct tcppacket *) pkt)->tcp.dest), ((struct tcppacket *) pkt)->ip.saddr);
      if (!*username)
	strcpy(username, UNKNOWN_IDENT_VALUE);
      log.f(LOGLEVEL, "%s connection attempt from %s@%s", service, username, remote_host);
    } else
#endif
      log.f(LOGLEVEL, "%s connection attempt from %s", service, remote_host);
  }
#ifdef _MULTITHREAD_
  free(pkt);
#endif
  return NULL;
}

#ifdef _MULTITHREAD_
void threaded_log_tcppacket(struct tcppacket pkt) {
  pthread_attr_t attr_t;
  pthread_t t;
  __u8 *image;

  image = (__u8 *) malloc(sizeof(struct tcppacket));
  memcpy(image, &pkt, sizeof(struct tcppacket));
  pthread_attr_init(&attr_t);
  pthread_attr_setdetachstate(&attr_t, PTHREAD_CREATE_DETACHED);
  pthread_create(&t, &attr_t, log_tcppacket, image);
  pthread_attr_destroy(&attr_t);
}
#endif

/*
 * log_tcp
 *
 * Main thread logging TCP connexions
 *
 */

void *log_tcp(void *nobody) {
  int s;
  struct tcppacket pkt;

  s = socket(AF_INET, SOCK_RAW, TCP_PROTO);
  if (s <= 0)
    exit(1);

  setuid(((struct passwd *)nobody)->pw_uid);

  for(;;) {
    if (read(s, (struct tcppacket *) &pkt, sizeof(struct tcppacket)) == -1)
      exit(1);

    if (pkt.ip.ihl != 5) {
#ifdef _MULTITHREAD_
      threaded_log_suspicious((struct iphdr *) &pkt);
#else
      log_suspicious((void *) &pkt);
#endif
      continue;
    }

    if (pkt.tcp.syn == 1 && pkt.tcp.ack == 0) {
#ifdef _MULTITHREAD_
      threaded_log_tcppacket(pkt);
#else
      log_tcppacket((void *) &pkt);
#endif
    }
  }
  /*@NOTREACHED@*/ 
  return NULL;
}
